from typing import Union, overload, Dict
import sympy

from mathpad.val import GenericVal, Val, Q
from mathpad.equation import Equation


@overload
def simplify(expr_or_eqn: GenericVal) -> GenericVal:
    ...


@overload
def simplify(expr_or_eqn: Equation) -> Equation:
    ...


def simplify(expr_or_eqn: Union[GenericVal, Equation]) -> Union[GenericVal, Equation]:
    if isinstance(expr_or_eqn, Equation):
        # TODO: simplification that actually makes use of equality
        return Equation(simplify(expr_or_eqn.lhs), simplify(expr_or_eqn.rhs))

    else:
        return expr_or_eqn.new(sympy.simplify(expr_or_eqn.val))


@overload
def factor(expr_or_eqn: GenericVal) -> GenericVal:
    ...


@overload
def factor(expr_or_eqn: Equation) -> Equation:
    ...


def factor(expr_or_eqn: Union[GenericVal, Equation]) -> Union[GenericVal, Equation]:
    if isinstance(expr_or_eqn, Equation):
        # TODO: simplification that actually makes use of equality
        return Equation(factor(expr_or_eqn.lhs), factor(expr_or_eqn.rhs))

    else:
        return expr_or_eqn.new(sympy.factor(expr_or_eqn.val))


@overload
def expand(expr_or_eqn: GenericVal) -> GenericVal:
    ...


@overload
def expand(expr_or_eqn: Equation) -> Equation:
    ...


def expand(expr_or_eqn: Union[GenericVal, Equation]) -> Union[GenericVal, Equation]:
    if isinstance(expr_or_eqn, Equation):
        # TODO: simplification that actually makes use of equality
        return Equation(expand(expr_or_eqn.lhs), expand(expr_or_eqn.rhs))

    else:
        return expr_or_eqn.new(sympy.expand(expr_or_eqn.val))


SubstitutionMap = Dict[Val, Q[Val]]


@overload
def subs(expr_or_eqn: GenericVal, substitutions: SubstitutionMap) -> GenericVal:
    ...


@overload
def subs(expr_or_eqn: Equation, substitutions: SubstitutionMap) -> Equation:
    ...


def subs(
    expr_or_eqn: Union[GenericVal, Equation],
    substitutions: SubstitutionMap,
) -> Union[GenericVal, Equation]:
    if isinstance(expr_or_eqn, Equation):
        return Equation(
            subs(expr_or_eqn.lhs, substitutions),
            subs(expr_or_eqn.rhs, substitutions),
        )

    else:
        sympy_subsmap = {}

        for from_, to in substitutions.items():
            if not isinstance(to, Val):
                to = from_.__class__(from_.units, to)

            sympy_subsmap[from_.val] = to.in_units(from_).val

        return expr_or_eqn.__class__(
            expr_or_eqn.units, expr_or_eqn.val.subs(sympy_subsmap)
        )



from typing import TYPE_CHECKING
import sympy
from sympy.physics.vector.printing import vlatex

if TYPE_CHECKING:
    from mathpad.val import GenericVal, Q


class Equation:
    def __init__(self, lhs: "Q[GenericVal]", rhs: "Q[GenericVal]"):
        from mathpad.val import Val

        lhs_is_pqty = isinstance(lhs, Val)
        rhs_is_pqty = isinstance(rhs, Val)

        assert lhs_is_pqty or rhs_is_pqty

        self.units = lhs.units if lhs_is_pqty else rhs.units  # type: ignore
        pqty_cls = (lhs if lhs_is_pqty else rhs).__class__

        if lhs_is_pqty and rhs_is_pqty:
            # rescale the rhs to match units
            rhs = rhs.in_units(lhs)
            # sanity check
            assert lhs.units == self.units == rhs.units

        self.lhs: Val = lhs if lhs_is_pqty else pqty_cls(self.units, rhs)
        self.rhs: Val = rhs if rhs_is_pqty else pqty_cls(self.units, rhs)

    def __repr__(self):
        return f"{self.lhs._repr(False)} = {self.rhs._repr(False)} {self.units}"

    def as_sympy_eq(self) -> sympy.Equality:
        return sympy.Equality(self.lhs.val, self.rhs.val)  # type: ignore

    # def as_units_eq_without_unknowns(self, unknowns: Collection[OutputVal]):
    #     return sympy.Equality(self.lhs.units, self.rhs.units)

    # Rich displays; Ipython etc
    def _repr_latex_(self):
        from mathpad.val import Val

        # use vlatex because it applies dot notation where possible
        lhs_ltx = (
            vlatex(self.lhs.val) if isinstance(self.lhs, Val) else str(self.lhs)
        ).replace("- 1.0 ", "-")

        rhs_ltx = (
            vlatex(self.rhs.val) if isinstance(self.rhs, Val) else str(self.rhs)
        ).replace("- 1.0 ", "-")

        units_ltx = "dimensionless" if self.units == 1 else vlatex(self.units)
        begin_center, end_center = "\\begin{center}", "\\end{center}"

        spacer_ltx = "\\hspace{1.25em}"

        return f"$$ {lhs_ltx} = {rhs_ltx} {spacer_ltx} {units_ltx} $$"



from mathpad import *


def kinetic_energy(m: Q[Mass], v: Q[Velocity]) -> Energy:
    "Energy of a moving object"
    return frac(1, 2) * m * v ** 2  # type: ignore


def elastic_energy(k: Q[Elasticity], dx: Q[Length]) -> Energy:
    return frac(1, 2) * k * dx ** 2  # type: ignore


def gravitational_energy(m: Q[Mass], h: Q[Length], g: Q[Acceleration] = g) -> Energy:
    return m * g * h  # type: ignore


def euler_lagrange(
    sum_kinetic_energy: Q[Energy],
    sum_potential_energy: Q[Energy],
    sum_non_conservative_forces: Q[Force],
    state: Val,
) -> Equation:
    L = sum_kinetic_energy - sum_potential_energy
    ds = diff(state)
    return diff(diff(L, wrt=ds)) - diff(L, wrt=state) == sum_non_conservative_forces


def impulse_momentum(
    m: Q[Mass],  # mass
    v1: Q[Velocity],  # initial velocity
    F: Q[Force],  # impulse force required
    t: Q[Time],  # impulse duration in seconds
    v2: Q[Velocity],  # final velocity
) -> Equation:
    "The force required in an instant to change an object's velocity"
    return m * v1 + integral(F, t) == m * v2


def velocity_acceleration(
    v: Q[Velocity], t: Q[Time]  # symbol for time
) -> Acceleration:
    return a == integral(v, t)


def force_momentum(
    m: Q[Mass],  # mass of object
    v: Q[Velocity],
    t: Q[Time] = t,
) -> Force:
    return diff(m * v)


def angular_momentum(
    r_p_o: Q[Length],  # unit vector of rotation axis (anti-clockwise)
    m: Q[Mass],  # mass of point object
    v: Q[Velocity],  # velocity of point object
) -> AngularMomentum:
    return r_p_o.cross(m * v)


# from typing import Annotated as An, get_type_hints

from mathpad import *


def resistance_resistivity(
    *,
    R: Q[Impedance],  # resistance
    rho: Q[Resistivity],  # resisitivity
    l: Q[Length],  # length of strip
    A: Q[Area]  # cross-sectional area of strip (usually very flat, but wide(ish))
):
    "Relate resistance and resistivity"
    return R == rho * l * A



from typing import Optional


class _GlobalOptions:
    auto_simplify: bool = True


_global_options = _GlobalOptions()


def set_global_options(auto_simplify: Optional[bool] = None):
    if auto_simplify is not None:
        _global_options.auto_simplify = auto_simplify



import sympy.physics.units.definitions as u
import sympy.physics.units.definitions.dimension_definitions as d

from mathpad.val import Unit, Dimensionless

# TODO: rearrange in alphabetical order


class Length(Unit):
    dimension = d.length


class Time(Unit):
    dimension = d.time


class AngularMomentum(Unit):
    dimension = d.mass * d.velocity * d.angle


class AngularVelocity(Unit):
    dimension = d.angle / d.time


class MomentOfInertia(Unit):
    dimension = d.mass / d.length ** 2


class Impedance(Unit):
    dimension = d.impedance


class Resistivity(Unit):
    dimension = d.impedance * d.length


class Inductance(Unit):
    dimension = d.inductance


class Capacitance(Unit):
    dimension = d.capacitance


class Mass(Unit):
    dimension = d.mass


# todo: figure out how to merge all Angle classes into one


class Angle(Dimensionless):
    # TODO: get deg, angular_mil and (maybe) steradian to use angle
    dimension = d.angle


class AngularMil(Angle):
    # has its own dimension for some reason
    dimension = u.angular_mil.dimension


class SteRadian(Angle):
    # this one's interesting because it's something like m**2 / m**2
    # subclass 'angle' but use its special dimension
    # TODO: review if this is gonna cause issues
    dimension = u.steradian.dimension


class Frequency(Unit):
    dimension = d.frequency


class Current(Unit):
    dimension = d.current


class Action(Unit):
    dimension = d.action


class AmountOfSubstance(Unit):
    dimension = d.amount_of_substance


class Radioactivity(Unit):
    # same SI unit as frequency but is (should be?) treated differently
    dimension = 1 / d.time


class Charge(Unit):
    dimension = d.charge


class Dioptre(Unit):
    # technicaly "Optical/Refractive Power" but this is more obvious
    dimension = 1 / d.length


class Lumosity(Unit):
    dimension = d.luminous_intensity


class GravityConstant(Unit):
    # useful to name this as a type; but is only ever a constant (defined in constants.py)
    dimension = d.length ** 3 / (d.mass * d.time ** 2)


class Gray(Unit):
    # technicaly measurement of "Absorbed Dose" but this is more obvious
    dimension = d.energy / d.mass


class Energy(Unit):
    dimension = d.energy


class Temperature(Unit):
    dimension = d.temperature


class Force(Unit):
    dimension = d.force


class Elasticity(Unit):
    dimension = d.force / d.length


class Pressure(Unit):
    dimension = d.pressure


class MolarGasConstant(Unit):
    # useful to name this as a specific dimension; but is only ever a constant (defined in constants.py)
    dimension = d.energy / (d.amount_of_substance * d.temperature)


class Conductance(Unit):
    dimension = d.conductance


class MagneticDensity(Unit):
    dimension = d.magnetic_density


class MagneticFlux(Unit):
    dimension = d.magnetic_flux


class Voltage(Unit):
    dimension = d.voltage


class Power(Unit):
    dimension = d.power


class Acceleration(Unit):
    dimension = d.acceleration


class Information(Unit):
    dimension = d.information


class Velocity(Unit):
    dimension = d.velocity


class LuminousIntensity(Unit):
    dimension = d.luminous_intensity


class Area(Unit):
    dimension = d.length ** 2


class Volume(Unit):
    dimension = d.length ** 3


class Permittivity(Unit):
    dimension = d.capacitance / d.length


class Katal(Unit):
    # Otherwise 'catalytic activity'
    dimension = d.amount_of_substance / d.time


class Illuminance(Unit):
    dimension = d.luminous_intensity / d.length ** 2


class Density(Unit):
    dimension = d.mass / d.length ** 3


class EnergyDensity(Unit):
    dimension = d.energy / d.length ** 3


class PlanckIntensity(Unit):
    # TODO; can/should we merge this with LuminousIntensity?
    dimension = d.mass / d.time ** 3


class Momentum(Unit):
    dimension = d.mass * d.velocity



from typing import Collection, Set, List, Optional, Tuple
from itertools import zip_longest

import sympy
import numpy as np
from sympy.core.function import Function, AppliedUndef
from sympy import Derivative
import plotly.graph_objects as go
from sympy.utilities.lambdify import lambdify
from scipy.integrate import RK45
from IPython.display import display
import plotly.io as pio

from mathpad.val import Val
from mathpad.equation import Equation
from mathpad.algebra import subs, SubstitutionMap, simplify
from mathpad._quality_of_life import t
from tqdm.notebook import tqdm


def simulate_dynamic_system(
    dynamics_equations: Collection[Equation],
    *,
    x_f: float,
    initial_conditions: SubstitutionMap,
    record: List[Val],
    max_step: Optional[float],
    substitute: SubstitutionMap = {},
    x_axis: Val = t,
    all_solutions: bool = False,
    # output display options
    verbose: bool = True,
    display_plots: bool = True,
    display_progress_bar: bool = True,
    display_explanation: bool = False,
    # plot formatting (display_plots=True)
    plot_static: bool = False,
    plot_static_figsize: Tuple[int, int] = (960, 400),
    plot_title: str = "Solution #{solutionNo}",
    _NEW_SOLVE: bool = False # TODO: fix this properly
) -> List[List[Tuple[float, List[float]]]]:
    "simulates a differential system specified by dynamics_equations from initial conditions at x_axis=0 (typically t=0) to x_final"

    verbose = verbose or display_explanation

    if plot_static_figsize and not plot_static:
        print(
            "Warning: plot_static_figsize was set but plot_static was not enabled. Enabling for you automatically. Set plot_static=True to "
        )

    # make static renderings a certain size, the default one is too square for my liking
    svg_renderer = pio.renderers["svg"]
    width, height = plot_static_figsize
    svg_renderer.width = width
    svg_renderer.height = height

    # TODO: support plotting on separate axes, and subplots
    # TODO: ensure we aren't subbing out something that is required for a 'record' output

    # TODO: support integrals
    if max_step is None:
        max_step = float("inf")

    if display_explanation:
        print("Using Input Equations:")
        for eqn in dynamics_equations:
            display(eqn)

        print("Subbing in Values:")
        for replace, _with in substitute.items():
            display(replace == _with)

    # pre-substitute and simplify the input equations before further processing
    problem_eqns = [simplify(subs(eqn, substitute)) for eqn in dynamics_equations]

    # collect derivatives and any unspecified unkowns
    derivatives: Set[Tuple[Function, float]] = set()

    for eqn in problem_eqns:
        sympy_eqn = eqn.as_sympy_eq()
        # TODO: properly check x_axis for derivative collection (usually t)
        derivatives.update(
            {
                (d.args[0], d.args[1][1] if isinstance(d.args[1], sympy.Tuple) else 1)
                for d in sympy_eqn.atoms(Derivative)
            }
        )
        derivatives.update(
            {(f, 0) for f in sympy_eqn.atoms(Function) if isinstance(f, AppliedUndef)}
        )

    highest_derivatives = {}
    lowest_derivatives = {}
    for f, lvl in derivatives:

        if f in highest_derivatives:
            if highest_derivatives[f] < lvl:
                highest_derivatives[f] = lvl
        else:
            highest_derivatives[f] = lvl

        if f in lowest_derivatives:
            if lowest_derivatives[f] > lvl:
                lowest_derivatives[f] = lvl
        else:
            lowest_derivatives[f] = lvl

    solve_for_highest_derivatives = [
        fn if lvl == 0 else sympy.diff(fn, (x_axis.val, lvl))
        for fn, lvl in highest_derivatives.items()
        if not lvl == lowest_derivatives[fn]
    ]

    solve_for_recorded_data = [val.val for val in record]

    solve_for = solve_for_highest_derivatives + solve_for_recorded_data

    _print_if(verbose, f"Solving subbed Equations{':' if display_explanation else ''}")

    if display_explanation:
        for eqn in problem_eqns:
            display(eqn)

        print("For values:")
        display(solve_for)

    solutions = sympy.solve(
        [eqn.as_sympy_eq() for eqn in problem_eqns],
        solve_for if _NEW_SOLVE else solve_for_highest_derivatives,
        dict=True,
    )

    assert any(solutions), "No Solution Found"
    _print_if(verbose, "Solving finished.")

    all_data = []

    for solution_idx, solution in enumerate(solutions):

        # in dict mode, if a solution is equal to the query the solution is not included in the dict
        # we need it down the line, so re-insert it here:
        for val in solve_for:
            if val not in solution:
                solution[val] = val

        # convert it to a vector for lambdify below
        solution_vec = [solution[val] for val in solve_for]

        if display_explanation:
            print(
                "Found Solution" + (f"#{solution_idx + 1}:" if all_solutions else ":")
            )
            for solve_val, result in zip(solve_for, solution_vec):
                eqn = sympy.Eq(solve_val, result)
                if eqn != True:  # this happens with passthrough variables
                    display(eqn)

        unknowns = set()
        for val in solution.values():
            unknowns.update(val.free_symbols)

        # since we're simulating along the x_axis, it doesn't count as an unknown here
        unknowns.remove(x_axis.val)

        assert not any(
            unknowns
        ), f"Cannot simulate ODE in the prescence of unknowns: {unknowns}. Please include them in substitutions"

        input_unzipped = []

        n_unique_derivatives = 0

        for fn, lowest_lvl in lowest_derivatives.items():
            highest_lvl = highest_derivatives[fn]

            if lowest_lvl == highest_lvl:
                continue
            
            n_unique_derivatives += 1

            input_unzipped.append(
                [
                    fn if lvl == 0 else sympy.diff(fn, (x_axis.val, lvl))
                    for lvl in range(lowest_lvl, highest_lvl)
                ]
            )

        # inputs are lowest to highest derivatives excluding the highest, ie [x, y, dx, dy]
        inputs = [
            deriv
            for derivatives in zip_longest(*input_unzipped)
            for deriv in derivatives
            if deriv is not None
        ]

        # outputs are highest of input derviatives plus recorded data
        # ie [dddx, record[0], record[1]]
        lambdified = lambdify([x_axis.val, inputs], solution_vec)

        data = []

        # define this integration function to record data and let outputs = diff(inputs)
        def step(x: float, state: np.ndarray):

            output = lambdified(x, state)

            highest_derivatives, recorded_data = (
                output[:n_unique_derivatives],
                output[n_unique_derivatives:],
            )

            data.append((x, recorded_data))

            dstate = np.append(state[n_unique_derivatives:], highest_derivatives)
            return dstate

        # normalize and check initial conditions
        y0 = []
        for sym in inputs:
            sym_hash = hash(sym)
            # find the original val
            # this is backwards, and this whole function could probably use a refactor
            for val in initial_conditions.keys():
                if hash(val) == sym_hash:  # because (hash(val) == hash(val.val))
                    break
            else:
                assert (
                    sym in initial_conditions
                ), f"Required initial condition missing: {sym}"

            replacement = initial_conditions[val]
            if isinstance(replacement, Val):
                val = float(replacement.in_units(val).val)
            else:
                val = replacement

            y0.append(val)

        integrator = RK45(step, t0=0, y0=y0, t_bound=x_f, max_step=max_step)

        _print_if(
            verbose,
            f"Simulating from t=0 to t={x_f} with a max_step of {max_step}"
            + (" with initial conditions:" if display_explanation else "."),
        )

        if display_explanation:
            for replace, _with in initial_conditions.items():
                display(replace == _with)

        t_prev = 0

        with tqdm(total=x_f, leave=False) if display_progress_bar else None as pbar:
            while integrator.status == "running":
                msg = integrator.step()

                if integrator.status == "failed":
                    print(f"integration completed with failed status: {msg}")
                    break

                if pbar:
                    dt = integrator.t - t_prev
                    pbar.update(dt)
                    t_prev = integrator.t

        _print_if(verbose, "Simulation finished. Plotting...")

        if display_plots:
            go.Figure(
                [
                    go.Scatter(
                        x=[t for t, _ in data],
                        y=[frame[idx] for _, frame in data],
                        name=str(sym),
                    )
                    for idx, sym in enumerate(record)
                ],
                layout={
                    "xaxis": {"title": str(x_axis)},
                    "title": {
                        "text": plot_title.format(solutionNo=solution_idx + 1),
                        "y": 0.9,
                        "x": 0.5,
                        "xanchor": "center",
                        "yanchor": "top",
                    },
                },
            ).show("svg" if plot_static else None)

        all_data.extend(data)

        if not all_solutions:
            break

    return all_data


def _print_if(condition: bool, msg: str):
    if condition:
        print(msg)


import sympy

from mathpad.val import *
from mathpad.units import *

from mathpad.units import *
from mathpad.solve import solve, Solution
from mathpad.equation import Equation
from mathpad._quality_of_life import t, g, pi, frac
from mathpad.algebra import subs, simplify, factor, expand
from mathpad.display import tabulate

from mathpad.functions import piecewise, sqrt
from mathpad.calculus import diff, integral
from mathpad.trigonometry import cos, sin, tan, magnitude, hypotenuse
from mathpad.simulate_dynamic_system import simulate_dynamic_system
from mathpad.vec3 import Vec3

try:
    from IPython.display import display
except ImportError:
    pass


sympy.init_printing()  # type: ignore
sympy.printing.printer.Printer.set_global_settings(min=-3, max=4)  # type: ignore

__version__ = "0.1.9"


from re import L
import sympy
from mathpad import *


def sin(x: Q[Angle]) -> Dimensionless:
    return Dimensionless(1, sympy.sin(x.val))  # type: ignore


def cos(x: Q[Angle]) -> Dimensionless:
    return Dimensionless(1, sympy.cos(x.val))  # type: ignore


def tan(x: Q[Angle]) -> Dimensionless:
    return Dimensionless(1, sympy.tan(x.val))  # type: ignore


def magnitude(*xs: Q[GenericVal]) -> GenericVal:
    "x's are lengths in orthogonal directions. ie (i, j k)"
    c_2 = 0
    for x in xs:
        c_2 += x ** 2
    return sqrt(c_2)


def hypotenuse(a: Q[Length], b: Q[Length]) -> Length:
    return cartesian_distance(a, b)


# TODO: add more trig functions


def sine_rule(
    a: Q[Length],  # length of side a
    b: Q[Length],  # length of side b
    alpha: Q[Angle],  # internal angle opposite to side a
    beta: Q[Angle],  # internal angle opposite to side b
) -> Equation:
    "Relates lengths of two sides of any triangle the internal angle opposite"
    return a == b * sin(alpha) / sin(beta)


def cosine_rule(
    a: Q[Length],  # length of side a
    b: Q[Length],  # length of side b
    c: Q[Length],  # length of side c
    C: Q[Angle],  # internal angle opposite to side c
) -> Equation:
    return c ** 2 == a ** 2 + b ** 2 - 2 * a * b * cos(C)



from typing import Collection, Dict, Set, Tuple, Union, overload
from ansitable import ANSITable
import sympy

from mathpad.val import Val, GenericVal
from mathpad.equation import Equation


class Solution:
    def __init__(self, result_dict: Dict[Val, Val]):
        self.result_dict = result_dict

    def __getitem__(self, k: GenericVal) -> GenericVal:
        result = self.result_dict[k]
        assert result.units == k.units
        return result

    def __repr__(self):
        return self._repr("", " ")

    def _repr(self, newline: str, indent: str):
        # TODO: print latex
        return (
            f"Solution({newline}{indent}"
            + f",{newline}{indent}".join(
                f"{key.val} = {val}" for key, val in self.result_dict.items()
            )
            + f"{newline}{' ' if not newline else ''})"
        )

    def print(self):
        table = ANSITable(
            "  unknown  ", "  solutions  ", border="thick", bordercolor="green"
        )
        for k, v in self.result_dict.items():
            table.row(k, v)
        table.print()


@overload
def solve(equations: Equation, solve_for: Val, in_place: bool = False) -> Solution:
    ...


@overload
def solve(
    equations: Collection[Equation],
    solve_for: Collection[Val],
    in_place: bool = False,
) -> Solution:
    ...


def solve(
    equations: Union[Equation, Collection[Equation]],
    solve_for: Union[Val, Collection[Val]],
    in_place: bool = False,
    # domain: Literal["complex", "real", "integers", "naturals", "naturals0"] = "real",
) -> Solution:
    # normalize inputs
    if isinstance(solve_for, Val):
        solve_for = (solve_for,)
    if isinstance(equations, Equation):
        equations = (equations,)

    assert len(equations) >= len(
        solve_for
    ), "Systems with less equations than unknowns are typically unsolveable"

    ukwn_syms = [ukwn.val for ukwn in solve_for]
    val_eqns = [eqn.as_sympy_eq() for eqn in equations]
    # units_eqns = [eqn.as_units_eq_without_unknowns(unknowns) for eqn in equations]

    # TODO: allow arbitrary expressions to be used here and substitute symbols for the user
    for sym in ukwn_syms:
        assert isinstance(sym, sympy.Expr)

    # treat integrals and derivatives like

    # TODO: add more domains. Are there more?
    # domain_set = {
    #     "real": S.Reals,
    #     "complex": S.Complexes,
    #     "integers": S.Integers,
    #     "naturals": S.Naturals0,
    #     "naturals0": S.Naturals0,
    # }

    results: Dict[sympy.Expr, sympy.Expr] = sympy.solve(val_eqns, ukwn_syms)  # type: ignore

    if not any(results):
        raise Exception("Solving failed!")

    solution = Solution(
        {unkwn: unkwn.__class__(unkwn.units, results[unkwn.val]) for unkwn in solve_for}
    )  # type: ignore

    if in_place:
        for unknown, result in solution.result_dict.items():
            unknown.val = result.val

    return solution


from typing import Any, Tuple, Type, Union, TypeVar, overload, Callable
import re
from abc import ABC

import sympy
import sympy.physics.units as su
from sympy.physics.units.dimensions import Dimension
from sympy.physics.units.systems.si import dimsys_SI
from sympy.physics.units.unitsystem import UnitSystem
from sympy.physics.vector import dynamicsymbols
from sympy.physics.vector.printing import vlatex
from sympy.physics.units.util import (
    quantity_simplify,
    convert_to,
)

from mathpad.equation import Equation
from mathpad.global_options import _global_options

# TODO: support numpy arrays
Num = Union[int, float, complex]
_UNSET = ()

# this should be a classmethod, but it isn't
_units2dimensional_expr = UnitSystem.get_default_unit_system().get_dimensional_expr


class Val(ABC):
    "An value with a set of units. For example 10 ohms or 20 meters / second**2"
    # TODO: documentation

    dimension: su.Dimension = _UNSET  # type: ignore

    def __new__(
        cls,
        units: su.Quantity,  # may also be a sympy expression of su.Quantities, ie su.meter**2
        val: Union[sympy.Expr, Num] = 1,  # gets set in __init__
    ):
        return super().__new__(cls)

    def __init_subclass__(cls):
        assert str(cls) == "<class 'mathpad.val.Unit'>" or cls.dimension is not _UNSET
        super().__init_subclass__()

    def new(
        self: "GenericVal", val: sympy.Expr, units: sympy.Expr = None
    ) -> "GenericVal":
        "Creates a new Val of the same class. if units == None; units = self.units"
        if units == None:
            units = self.units
        return self.__class__(units, val)

    def __init__(
        self,
        units: su.Quantity,  # may also be a sympy expression of su.Quantities, ie su.meter**2
        val: Union[sympy.Expr, Num] = 1,
    ):

        self.val: sympy.Symbol = sympy.sympify(val)
        self.units: su.Quantity = quantity_simplify(sympy.sympify(units))

        assert self.dimension is not _UNSET
        units_dimension = _units2dimensional_expr(self.units)

        if _is_dimensionless(self.dimension) and _is_dimensionless(units_dimension):
            return

        else:
            assert dimsys_SI.equivalent_dims(units_dimension, self.dimension), (
                f"Units {self.units} do not match the dimensionality of {self.__class__.__name__} ({self.dimension}).\n"
                f"Instead got {units_dimension}"
            )

    def __hash__(self):
        return hash(self.val)

    @overload
    def __eq__(self, other: "OutputVal") -> "Equation":
        ...

    @overload
    def __eq__(self: "GenericVal", other: "Q[GenericVal]") -> "Equation":
        ...

    def __eq__(self, other: "Q[Val]") -> "Equation":
        if isinstance(other, Val):
            SumDimensionsMismatch.check(self, "==", other)
        return Equation(self, other)

    def __req__(self, other: Num) -> "Equation":
        return Equation(other, self)

    def __repr__(self) -> str:
        return self._repr(True)

    # Rich displays; Ipython etc
    def _repr_latex_(self):
        # use vlatex because it applies dot notation where possible
        val_ltx = vlatex(self.val)
        clean_val_ltx = val_ltx.replace("- 1.0 ", "-")
        units_ltx = "dimensionless" if self.units == 1 else vlatex(self.units)

        begin_center, end_center = "\\begin{center}", "\\end{center}"
        begin_center, end_center = "", ""

        spacer_ltx = "\\hspace{1.25em}"
        # remove '1.0's

        return f"$$ {begin_center} {clean_val_ltx} {spacer_ltx} {units_ltx} {end_center} $$"

    def _repr_png_(self):
        return self.val._repr_png_()

    def _repr_svg_(self):
        return self.val._repr_svg_()

    def _repr_disabled(self):
        return self.val._repr_disabled()

    def _repr(self, with_units: bool) -> str:
        # ignore 0's following last non-zero decimal
        res = re.sub(
            r"(\d+\.\d*[^0])0+",
            r"\1",
            str(self.val if _is_primitive_num(self.val) else self.val.evalf(6)),  # type: ignore
        )
        res = re.sub(r"\.0+([e\*]|$)", r"\1", res)

        # also ignore 1*x
        res = re.sub(r"1\*", "", res)

        # replace "Derivative" and "Integral" with their input equivalents

        # TODO: Make derivatives and integrals easier to read
        # res = res.replace("Derivative", "ᵈ⁄dt").replace("Integral", "∫")

        # do surgery to get display how we want it with very specific rules
        # not going to work for complex expressions TODO: make it

        if with_units and self.units != 1:
            units_str = re.sub(
                r"(.*?[a-zA-Z])(\/|\*\*|$)", r"\1s\2", str(self.units), 1
            )

            res += f" {units_str}"

        # TODO: use superscript for exponents (both val and units)

        return res

    def in_units(self: "GenericVal", units: Union[str, "Val"]) -> "GenericVal":
        if isinstance(units, str):
            if units == "si":
                units = "SI"
            new_units = UnitSystem.get_unit_system(units)._base_units

        else:
            SumDimensionsMismatch.check(self, ".in_units", units)
            new_units = units.units

        units_factor, new_units = _split_coeff_and_units(
            convert_to(self.units, new_units)  # type: ignore
        )
        new_val = units_factor * self.val

        return self.new(new_val, new_units)

    @overload
    def __add__(self, other: "OutputVal") -> "OutputVal":
        ...

    @overload
    def __add__(self: "GenericVal", other: Num) -> "GenericVal":
        ...

    @overload
    def __add__(self: "GenericVal", other: Num) -> "GenericVal":
        ...

    @overload
    def __add__(self: "GenericVal", other: "GenericVal") -> "GenericVal":
        ...

    def __add__(self, other: "Q[GenericVal]"):
        return self._sum_op(other, lambda a, b: a + b, "+", False)

    def __radd__(self: "GenericVal", other: Num) -> "GenericVal":
        return self._sum_op(other, lambda a, b: b + a, "+", True)

    @overload
    def __sub__(self: "GenericVal", other: "OutputVal") -> "GenericVal":
        ...

    @overload
    def __sub__(self, other: "GenericVal") -> "GenericVal":
        ...

    @overload
    def __sub__(self: "GenericVal", other: Num) -> "GenericVal":
        ...

    @overload
    def __sub__(self: "GenericVal", other: "GenericVal") -> "GenericVal":
        ...

    def __sub__(self, other: "Q[GenericVal]"):
        return self._sum_op(other, lambda a, b: a - b, "-", False)

    def __neg__(self):
        return self.new(-self.val)  # type: ignore

    def __rsub__(self: "GenericVal", other: Num) -> "GenericVal":
        return self._sum_op(other, lambda a, b: b - a, "-", True)

    @overload
    def __mul__(self, other: "Val") -> "OutputVal":
        ...

    @overload
    def __mul__(self: "GenericVal", other: Num) -> "GenericVal":
        ...

    def __mul__(self, other: "Q[Val]") -> "OutputVal":
        return self._prod_op(other, lambda a, b: a * b, is_pow=False)

    def __call__(self, x) -> "Val":
        return self.new(self.val(x))

    # because matrix multiplication has its own operator, multiplication can be commutative
    # sympy also reorders everything under the hood (alphabetically, I believe; so preserving order is pointless)

    def __rmul__(self: "GenericVal", other: Union[Num, str]) -> "GenericVal":

        if isinstance(other, str):
            assert self.val == 1, "Attempted to create variable with a non-unit Val"

            # TODO: support variables which are functions of a symbol other than t. should this be a fn() fn?
            if "(" in other:
                assert (
                    other.count("(") == 1 and other.count(")") == 1 and other[-1] == ")"
                ), f"Malformed variable name. Variables which are functions of symbols must take the form 'f(x)'. Insted got {other}"

                function_name, the_rest = other.split("(")
                function_of = [x.strip() for x in the_rest[:-1].split(",")]
                if len(function_of) == 1 and function_of[0] == "t":
                    sym = dynamicsymbols(function_name)

                else:
                    # TODO: test this
                    sym = sympy.Function(function_name)(function_of)
            else:
                sym = sympy.Symbol(other)

            return self.new(sym)

        else:
            return self._prod_op(other, lambda a, b: b * a, is_pow=False)

    @overload
    def __truediv__(self, other: "Val") -> "OutputVal":
        ...

    @overload
    def __truediv__(self: "GenericVal", other: Num) -> "GenericVal":
        ...

    def __truediv__(self, other: "Q[Val]") -> "OutputVal":
        return self._prod_op(other, lambda a, b: a / b, is_pow=False)

    def __rtruediv__(self, other: Num) -> "OutputVal":
        return self._prod_op(other, lambda a, b: b / a, is_pow=False)

    # exponents should always be dimensionless
    def __pow__(self, other: "Q[Val]") -> "OutputVal":
        if isinstance(other, Val) and not _is_dimensionless(other.dimension):
            raise DimensionalExponentError(
                f"Exponents must always be dimensionless. Instead got {other.dimension}: {other}"
            )
        return self._prod_op(other, lambda a, b: a ** b, is_pow=True)

    __xor__ = __pow__

    def __rpow__(self, other: Num) -> "Dimensionless":
        if not _is_dimensionless(self.dimension):
            raise DimensionalExponentError(
                f"Exponents must always be dimensionless. Instead got {self.dimension}: {self}"
            )
        res = self._prod_op(other, lambda a, b: b ** a, is_pow=True)
        assert isinstance(res, Dimensionless)
        return res

    __rxor__ = __pow__

    def _sum_op(
        self: "GenericVal",
        other: "Q[GenericVal]",
        op: Callable[[Any, Any], Any],
        op_str: str,
        reverse: bool,
    ) -> "GenericVal":
        other_units, other_val = (
            (other.units, other.val) if isinstance(other, Val) else (self.units, other)
        )

        if isinstance(other, Val):
            SumDimensionsMismatch.check(
                other if reverse else self, op_str, self if reverse else other
            )

        other_units_rescale_factor = quantity_simplify(
            convert_to(other_units, self.units) / self.units
        )

        # choose the larger of the two input units as the output units
        use_other_units = other_units_rescale_factor > 1
        new_units = other_units if use_other_units else self.units

        self_val_rescaled = (
            self.val / other_units_rescale_factor if use_other_units else self.val
        )
        other_val_rescaled = (
            other_val if use_other_units else other_val * other_units_rescale_factor
        )

        new_val = op(self_val_rescaled, other_val_rescaled)

        res = self.new(new_val, new_units)

        if _global_options.auto_simplify:
            from mathpad.algebra import simplify

            return simplify(res)

        else:
            return res

    def _prod_op(
        self,
        other: "Q[Val]",
        op: Callable[[Any, Any], Any],
        is_pow: bool,
    ) -> "OutputVal":
        other_units, other_val = (
            (other.units, other.val)
            if isinstance(other, Val)
            else (other if is_pow else 1, other)
        )

        rescale_factor, new_units = _split_coeff_and_units(op(self.units, other_units))

        new_val = rescale_factor * op(self.val, other_val)
        if isinstance(new_val, sympy.Expr):
            new_val = quantity_simplify(new_val)

        dimension_unchanged = dimsys_SI.equivalent_dims(
            self.dimension, _units2dimensional_expr(new_units)
        )

        if dimension_unchanged:
            res = self.new(new_val, new_units)

        elif new_units == Dimensionless.default_units:
            res = Dimensionless(new_units, new_val)  # type: ignore

        else:
            # TODO: a big lookup table for dimensional expr -> Val subclass
            res = OutputVal(new_units, new_val)  # type: ignore

        if _global_options.auto_simplify:
            from mathpad.algebra import simplify

            return simplify(res)

        else:
            return res


GenericVal = TypeVar("GenericVal", bound=Val)
Q = Union[GenericVal, Num, "OutputVal"]


class DimensionError(TypeError):
    pass


class DimensionalExponentError(DimensionError):
    pass


class SumDimensionsMismatch(DimensionError):
    def __init__(
        self,
        left: Val,
        operation: str,
        right: Val,
        extra_msg: str = "",
    ):
        self.args = (
            f'{extra_msg}[{left.dimension}: {left}] "{operation}" [{right.dimension}: {right}]',
        )

    @classmethod
    def check(cls, a: Val, op_str: str, b: Val):
        if not dimsys_SI.equivalent_dims(a.dimension, b.dimension) and not (
            _is_dimensionless(a.dimension) and _is_dimensionless(b.dimension)
        ):
            raise cls(a, op_str, b)


def _split_coeff_and_units(unit_expr: sympy.Expr) -> Tuple[float, Any]:
    # TODO: test and refactor
    converted = quantity_simplify(unit_expr)  # type: ignore
    if converted.is_Number:
        return converted, Dimensionless.default_units  # type: ignore

    try:
        units_factor = converted.args[0]
        if units_factor.is_Number:
            return units_factor, quantity_simplify(unit_expr / units_factor)

    except AttributeError:
        pass

    return 1, unit_expr


def _is_primitive_num(x: Q[Val]):
    return isinstance(x, (int, float, complex))


class Unit(Val):
    pass


class Dimensionless(Unit):

    dimension = su.Dimension(1)  # type: ignore
    default_units = sympy.sympify(1)

    def __new__(
        cls,
        units: su.Quantity,  # may also be a sympy expression of su.Quantities, ie su.meter**2
        val: Union[sympy.Expr, Num] = 1,  # gets set in __init__
    ):
        # measures of angle are technically dimensionless
        assert _is_dimensionless(_units2dimensional_expr(units))
        self = super().__new__(cls, units, val)
        return self


class OutputVal(Val):
    """A physical quantity whose dimensionality cannot be type-hinted, but will still check at runtime."""

    dimension = None  # type: ignore # gets set in __init__

    def __new__(
        cls,
        units: su.Quantity,  # may also be a sympy expression of su.Quantities, ie su.meter**2
        val: Union[sympy.Expr, Num] = 1,  # gets set in __init__
    ):
        dimension = _units2dimensional_expr(units)

        # catch the case when a quantity is dimensionless (ie 1 meter/meter) -
        # class as Dimensionless for
        if _is_dimensionless(dimension):
            res = Dimensionless(units, val)

        else:
            res = super().__new__(cls, units, val)
            res.dimension = dimension  # type: ignore

        return res


def _is_dimensionless(dimension):
    from mathpad.dimensions import Angle, AngularMil, SteRadian

    # accepts output of units2dimensional_expr (an expression, not a dimension object); so normalize here
    if isinstance(dimension, Dimension):
        dimension = dimension.args[0]

    # TODO: clean up this mess of a function
    return (
        any(
            dimension == dim
            for dim in [
                1,
                Angle.dimension.args[0],
                Dimensionless.dimension.args[0],
                SteRadian.dimension.args[0],
                AngularMil.dimension.args[0],
            ]
        )
        or dimsys_SI.get_dimensional_dependencies(dimension) == {}
    )



import os

import sympy.physics.units.definitions as u
from sympy.physics.units.quantities import Quantity
from sympy.physics.units.systems.si import dimsys_SI

from mathpad.val import Val
import mathpad.dimensions as dims

HERE = os.path.dirname(__file__)

AUTOGEN_WARNING = '"WARNING: this file was automatically generated by commented code within mathpad.py, and then modified slightly"\n'

print("writing to " + f"{HERE}/units.py\nand {HERE}/constants.py")

with open(f"{HERE}/units.py", "w") as units_f, open(
    f"{HERE}/constants.py", "w"
) as constants_f:

    # top of 'units.py'
    units_f.write(AUTOGEN_WARNING)
    units_f.write("import sympy.physics.units.definitions.unit_definitions as u\n")
    units_f.write("from mathpad.physical_quantities import *\n")

    # top of 'constants.py'
    constants_f.write(AUTOGEN_WARNING)
    constants_f.write("import sympy.physics.units.definitions.unit_definitions as u\n")
    constants_f.write("from mathpad.physical_quantities import *\n")

    for name in dir(u):
        qty = getattr(u, name)

        if name.endswith("constant"):
            constants_f.write(
                f"""
{name} = OutputVal(u.{name})"""
            )

        elif isinstance(qty, Quantity):

            for quantity_cls_name in dir(dims):
                if quantity_cls_name in "du":
                    continue

                quantity_cls = getattr(dims, quantity_cls_name)

                # handle the Angle dimension specially
                if str(qty).startswith("angular") or str(qty).startswith("rad"):
                    units_f.write(
                        f"""
{name} = Angle(u.{name})"""
                    )
                    break

                elif (
                    isinstance(quantity_cls, type)
                    and issubclass(quantity_cls, Val)
                    and quantity_cls is not Val
                    and dimsys_SI.equivalent_dims(qty.dimension, quantity_cls.dimension)
                ):
                    units_f.write(
                        f"""
{name} = {quantity_cls_name}(u.{name})"""
                    )
                    break
            else:
                print(f"Warning: Quantity '{name}' could not be matched; skipping...")



from typing import Generic, Sequence
from sympy.physics.vector import vlatex

from mathpad.val import GenericVal, Q, Val
from mathpad._quality_of_life import t
from mathpad.trigonometry import magnitude
from mathpad.calculus import diff
from mathpad.algebra import SubstitutionMap, subs

# TODO: use sympy.physics.vector.ReferenceFrame


class Vec3(Generic[GenericVal], Sequence):
    def __init__(
        self, *, i: Q[GenericVal] = 0, j: Q[GenericVal] = 0, k: Q[GenericVal] = 0
    ):
        "assumes meters if Num passed"
        set_units = None
        for attempt in [i, j, k]:
            try:
                attempt.units
                if set_units:
                    # do dimensionality check
                    # TODO: make error message better
                    i.in_units(attempt)
                set_units = attempt

            except AttributeError:
                pass

        assert set_units, "At least one input must be a Val with dimensionality"

        self.i: GenericVal = i if isinstance(i, Val) else set_units.new(i)  # type: ignore
        self.j: GenericVal = j if isinstance(j, Val) else set_units.new(j)  # type: ignore
        self.k: GenericVal = k if isinstance(k, Val) else set_units.new(k)  # type: ignore
        self._tuple = (self.i, self.j, self.k)

    def __getitem__(self, idx):
        return self._tuple[idx]

    def __iter__(self):
        return iter(self._tuple)

    def __len__(self):
        return 3
    
    def subs(self, substitutions: SubstitutionMap):
        return Vec3(
            i=subs(self.i, substitutions),
            j=subs(self.j, substitutions),
            k=subs(self.k, substitutions)
        )

    def dot(self, other: "Vec3"):
        ai, aj, ak = self
        bi, bj, bk = other
        return ai * bi + aj * bj + ak * bk  # type: ignore

    def _repr_latex_(self):
        # use vlatex because it applies dot notation where possible

        res = (
            "\\begin{bmatrix} "
            + " \\\\ ".join(
                f'{vlatex(x.val).replace("- 1.0 ", "-")} & {"dimensionless" if x.units == 1 else vlatex(x.units)}'
                + "\\cdot \\hat{"
                + unit_vector
                + "}"
                for unit_vector, x in zip("ijk", self)
            )
            + " \\end{bmatrix}"
        )

        return f"$$ {res} $$"

    def diff(self, n: int = 1, wrt: Val = t):
        i, j, k = diff(self, n=n, wrt=wrt)  # type: ignore
        return Vec3(i=i, j=j, k=k)

    def __add__(self: "Vec3[Val]", other: "Vec3[Val]"):
        ai, aj, ak = self
        bi, bj, bk = other
        return Vec3(
            i=ai + bi,  # type: ignore
            j=aj + bj,  # type: ignore
            k=ak + bk,  # type: ignore
        )

    def __mul__(self: "Vec3[GenericVal]", other: Val):
        ai, aj, ak = self
        return Vec3(
            i=ai * other,  # type: ignore
            j=aj * other,  # type: ignore
            k=ak * other,  # type: ignore
        )

    def __rmul__(self: "Vec3[GenericVal]", other: Val):
        return self * other  # call __mull__

    def __abs__(self):
        return magnitude(*self)

    def __neg__(self):
        return Vec3(
            i=-self.i,
            j=-self.j,
            k=-self.k,
        )


"WARNING: this file was automatically generated by commented code within mathpad.py, and then modified slightly"
import sympy.physics.units.definitions.unit_definitions as u
from mathpad.val import OutputVal

atomic_mass_constant = OutputVal(u.atomic_mass_constant, 1)
avogadro_constant = OutputVal(u.avogadro_constant, 1)
boltzmann_constant = OutputVal(u.boltzmann_constant, 1)
coulomb_constant = OutputVal(u.coulomb_constant, 1)
coulombs_constant = OutputVal(u.coulombs_constant, 1)
electric_constant = OutputVal(u.electric_constant, 1)
electric_force_constant = OutputVal(u.electric_force_constant, 1)
faraday_constant = OutputVal(u.faraday_constant, 1)
gravitational_constant = OutputVal(u.gravitational_constant, 1)
josephson_constant = OutputVal(u.josephson_constant, 1)
magnetic_constant = OutputVal(u.magnetic_constant, 1)
molar_gas_constant = OutputVal(u.molar_gas_constant, 1)
stefan_boltzmann_constant = OutputVal(u.stefan_boltzmann_constant, 1)
von_klitzing_constant = OutputVal(u.von_klitzing_constant, 1)


from typing import Iterable, List, Union, overload
from mathpad.val import Val, OutputVal, Q
from mathpad import t
from mathpad.global_options import _global_options
from mathpad.algebra import simplify
import sympy

# TODO: prescale values and units


@overload
def diff(val: Q[Val], n: int = 1, *, wrt: Val = t) -> OutputVal:
    ...


@overload
def diff(val: Iterable[Q[Val]], n: int = 1, *, wrt: Val = t) -> List[OutputVal]:
    ...


def diff(
    val: Union[Q[Val], Iterable[Q[Val]]], n: int = 1, *, wrt: Val = t
) -> Union[OutputVal, List[OutputVal]]:

    try:  # is val iterable? assume its a vector
        iter(val)
        return [diff(pq, wrt=wrt, n=n) for pq in val]

    except TypeError:
        res = OutputVal(
            val.units / wrt.units, val.val.diff((wrt.val, n))  # type: ignore
        )

        if _global_options.auto_simplify:
            res = simplify(res)

        return res


def integral(val: Val, wrt: Val = t) -> OutputVal:
    res = OutputVal(
        val.units * wrt.units, sympy.integrate(val.val, wrt.val)  # type: ignore
    )

    if _global_options.auto_simplify:
        res = simplify(res)

    return res



import sympy

from mathpad.units import seconds, meters, second, radians
from mathpad.dimensions import Angle

t = "t" * seconds
g = 9.81 * meters / second ** 2
pi = Angle(radians.units, sympy.pi)  # type: ignore


def frac(numerator, denominator):
    return sympy.Rational(numerator, denominator)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mathpad']

package_data = \
{'': ['*']}

install_requires = \
['ansitable>=0.9.6,<0.10.0',
 'numpy>=1.20,<2.0',
 'plotly>=5.3.1,<6.0.0',
 'scipy>=1.7.1,<2.0.0',
 'sympy>=1.8,<2.0',
 'tqdm>=4.62.3,<5.0.0',
 'typing-extensions>=3.10.0,<4.0.0']

setup_kwargs = {
    'name': 'mathpad',
    'version': '0.1.9',
    'description': 'Simplified interface to Sympy for solving physics, engineering and maths problems',
    'long_description': '# MathPad\n\n<!-- TODO: set up all the services needed for these badges -->\n<!-- <p align="center">\n  <a href="https://github.com/CallumJHays/mathpad/actions?query=workflow%3ACI">\n    <img src="https://img.shields.io/github/workflow/status/CallumJHays/mathpad/CI/main?label=CI&logo=github&style=flat-square" alt="CI Status" >\n  </a>\n  <a href="https://mathpad.readthedocs.io">\n    <img src="https://img.shields.io/readthedocs/mathpad.svg?logo=read-the-docs&logoColor=fff&style=flat-square" alt="Documentation Status">\n  </a>\n  <a href="https://codecov.io/gh/CallumJHays/mathpad">\n    <img src="https://img.shields.io/codecov/c/github/CallumJHays/mathpad.svg?logo=codecov&logoColor=fff&style=flat-square" alt="Test coverage percentage">\n  </a>\n</p>\n<p align="center">\n  <a href="https://python-poetry.org/">\n    <img src="https://img.shields.io/badge/packaging-poetry-299bd7?style=flat-square&logo=data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA4AAAASCAYAAABrXO8xAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAJJSURBVHgBfZLPa1NBEMe/s7tNXoxW1KJQKaUHkXhQvHgW6UHQQ09CBS/6V3hKc/AP8CqCrUcpmop3Cx48eDB4yEECjVQrlZb80CRN8t6OM/teagVxYZi38+Yz853dJbzoMV3MM8cJUcLMSUKIE8AzQ2PieZzFxEJOHMOgMQQ+dUgSAckNXhapU/NMhDSWLs1B24A8sO1xrN4NECkcAC9ASkiIJc6k5TRiUDPhnyMMdhKc+Zx19l6SgyeW76BEONY9exVQMzKExGKwwPsCzza7KGSSWRWEQhyEaDXp6ZHEr416ygbiKYOd7TEWvvcQIeusHYMJGhTwF9y7sGnSwaWyFAiyoxzqW0PM/RjghPxF2pWReAowTEXnDh0xgcLs8l2YQmOrj3N7ByiqEoH0cARs4u78WgAVkoEDIDoOi3AkcLOHU60RIg5wC4ZuTC7FaHKQm8Hq1fQuSOBvX/sodmNJSB5geaF5CPIkUeecdMxieoRO5jz9bheL6/tXjrwCyX/UYBUcjCaWHljx1xiX6z9xEjkYAzbGVnB8pvLmyXm9ep+W8CmsSHQQY77Zx1zboxAV0w7ybMhQmfqdmmw3nEp1I0Z+FGO6M8LZdoyZnuzzBdjISicKRnpxzI9fPb+0oYXsNdyi+d3h9bm9MWYHFtPeIZfLwzmFDKy1ai3p+PDls1Llz4yyFpferxjnyjJDSEy9CaCx5m2cJPerq6Xm34eTrZt3PqxYO1XOwDYZrFlH1fWnpU38Y9HRze3lj0vOujZcXKuuXm3jP+s3KbZVra7y2EAAAAAASUVORK5CYII=" alt="Poetry">\n  </a>\n  <a href="https://github.com/ambv/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square" alt="black">\n  </a>\n  <a href="https://github.com/pre-commit/pre-commit">\n    <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white&style=flat-square" alt="pre-commit">\n  </a>\n</p>\n<p align="center">\n  <a href="https://pypi.org/project/mathpad/">\n    <img src="https://img.shields.io/pypi/v/mathpad.svg?logo=python&logoColor=fff&style=flat-square" alt="PyPI Version">\n  </a>\n  <img src="https://img.shields.io/pypi/pyversions/mathpad.svg?style=flat-square&logo=python&amp;logoColor=fff" alt="Supported Python versions">\n  <img src="https://img.shields.io/pypi/l/mathpad.svg?style=flat-square" alt="License">\n</p> -->\n\nType-hinted, simplified interface to `sympy` for solving engineering, science and maths problems.\n\n## Example\n\nA car is driving at 5 miles per hour. The driver hits the brakes,\ndecelerating at a rate of 2 meters per second squared. How long will the car take to come to come to a halt?\n\n```python\nfrom mathpad import *\n\nt = "t" * seconds  # "t seconds"\n\nvelocity = 5 * miles / hour  # "5 miles/hour"\n\ndeceleration = 2 * m / s ** 2  # "2 meters/second**2"\n\nsolution = solve(t, t == velocity / deceleration)\nprint(solution) # "Solution(t = 1.1176 seconds)"\n```\n\nCheck out more examples in the [Examples directory](./examples/)\n\n## Installation\n\nInstall via pip:\n\n`pip install mathpad`\n\n<!-- ## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)): -->\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<!-- markdownlint-enable -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\n<!-- This project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome! -->\n\n## Credits\n\nThis package was created with\n[Cookiecutter](https://github.com/audreyr/cookiecutter) and the\n[browniebroke/cookiecutter-pypackage](https://github.com/browniebroke/cookiecutter-pypackage)\nproject template.\n',
    'author': 'Callum J Hays',
    'author_email': 'callumjhays@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/CallumJHays/mathpad',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from typing import Union
from ansitable import ANSITable
from mathpad.val import Val
from mathpad.equation import Equation


def tabulate(*entities: Union[Val, Equation]):
    "Prints a list of values or relations consistent with the display environment"

    # TODO: latex version in supporting IPython environments
    table = ANSITable("  Entity  ", "  Units  ", border="thick", bordercolor="blue")

    for entity in entities:
        if isinstance(entity, Val):
            table.row(entity.val, entity.units)

        else:
            table.row(f"{entity.lhs} == {entity.rhs}", entity.units)

    table.print()


"Note: Unfinished module!"
from typing import (
    Any,
    OrderedDict,
    Union,
)

from sympy.vector import CoordSys3D

from mathpad.val import (
    OutputVal,
)

R3 = CoordSys3D("ℝ3")


class VectorSpace(OutputVal):
    def __init__(self, ordered_units: OrderedDict[str, OutputVal]):
        self.ordered_units = ordered_units

    def __getattribute__(self, name: str) -> Any:
        return self.ordered_units.get(name)

    def __iter__(self):
        return self.ordered_units.values()

    def __getitem__(self, idx: Union[int, slice]):
        ords: OrderedDict[str, OutputVal] = self.ordered_units
        return list(ords.values())[idx]


# class Vector(OutputVal):
#     # TODO: review typing once PEP TypeVarTuple lands
#     def __new__(self, vector_space, val: Iterable[OutputVal]):
#         super().__init__(units, val)

# def vector_var(
#     name: str, vector_space: Union[VectorSpace, OrderedDict[str, OutputVal]]
# ) -> Vector:
#     sym = sympy.Symbol(name, **assumptions)
#     if not isinstance(vector_space, VectorSpace):
#         vector_space = VectorSpace(vector_space)
#     res =

#     return res



from typing import List, Optional, Tuple
from sympy import Piecewise

from mathpad.val import Val, GenericVal, Q
from mathpad._quality_of_life import frac

# TODO: improve API once ">", "<", ">=" etc operators are implemented for Val
def piecewise(x: Val, region_vals: List[Tuple[float, Q[GenericVal]]]) -> GenericVal:
    "a piecewise series of <"
    assert any(region_vals)
    assert region_vals[-1][0] == float("inf")
    inp = []
    prev_pqty: Optional[Val] = None
    for lt, pqty in region_vals:
        rescaled = pqty.in_units(prev_pqty) if prev_pqty is not None else pqty
        inp.append((rescaled.val, x.val < lt))
        prev_pqty = pqty

    return prev_pqty.new(Piecewise(*inp))


def sqrt(x: Q[Val]):
    return x ** frac(1, 2)


"WARNING: this file was automatically generated by commented code within mathpad.py, and then modified slightly"
import sympy.physics.units.definitions.unit_definitions as u
from mathpad.dimensions import *

A = Current(u.A)
Bq = Frequency(u.Bq)
C = Charge(u.C)
D = Dioptre(u.D)
F = Capacitance(u.F)
G = GravityConstant(u.G)
Gy = Gray(u.Gy)
H = Inductance(u.H)
Hz = Frequency(u.Hz)
J = Energy(u.J)
K = Temperature(u.K)
N = Force(u.N)
Pa = EnergyDensity(u.Pa)
R = MolarGasConstant(u.R)
S = Conductance(u.S)
T = MagneticDensity(u.T)
V = Voltage(u.V)
W = Power(u.W)
Wb = MagneticFlux(u.Wb)
Z0 = Impedance(u.Z0)
acceleration_due_to_gravity = Acceleration(u.acceleration_due_to_gravity)
ampere = Current(u.ampere)
amperes = Current(u.amperes)
amu = Mass(u.amu)
amus = Mass(u.amus)
rad = Angle(u.rad)
angular_mil = AngularMil(u.angular_mil)
angular_mils = AngularMil(u.angular_mils)
anomalistic_year = Time(u.anomalistic_year)
anomalistic_years = Time(u.anomalistic_years)
astronomical_unit = Length(u.astronomical_unit)
astronomical_units = Length(u.astronomical_units)
atm = EnergyDensity(u.atm)
atmosphere = EnergyDensity(u.atmosphere)
atmospheres = EnergyDensity(u.atmospheres)
atomic_mass_unit = Mass(u.atomic_mass_unit)
au = Length(u.au)
avogadro_number = Dimensionless(u.avogadro_number)
bar = EnergyDensity(u.bar)
bars = EnergyDensity(u.bars)
becquerel = Frequency(u.becquerel)
bit = Information(u.bit)
bits = Information(u.bits)
byte = Information(u.byte)
c = Velocity(u.c)
candela = LuminousIntensity(u.candela)
candelas = LuminousIntensity(u.candelas)
cd = LuminousIntensity(u.cd)
centiliter = Volume(u.centiliter)
centiliters = Volume(u.centiliters)
centimeter = Length(u.centimeter)
centimeters = Length(u.centimeters)
cl = Volume(u.cl)
cm = Length(u.cm)
common_year = Time(u.common_year)
common_years = Time(u.common_years)
coulomb = Charge(u.coulomb)
coulombs = Charge(u.coulombs)
curie = Frequency(u.curie)
day = Time(u.day)
days = Time(u.days)
deciliter = Volume(u.deciliter)
deciliters = Volume(u.deciliters)
decimeter = Length(u.decimeter)
decimeters = Length(u.decimeters)
deg = Dimensionless(u.deg)
degree = Dimensionless(u.degree)
degrees = Dimensionless(u.degrees)
dioptre = Dioptre(u.dioptre)
dl = Volume(u.dl)
dm = Length(u.dm)
draconic_year = Time(u.draconic_year)
draconic_years = Time(u.draconic_years)
e0 = Permittivity(u.e0)
eV = Energy(u.eV)
electronvolt = Energy(u.electronvolt)
electronvolts = Energy(u.electronvolts)
elementary_charge = Charge(u.elementary_charge)
exbibyte = Information(u.exbibyte)
exbibytes = Information(u.exbibytes)
farad = Capacitance(u.farad)
farads = Capacitance(u.farads)
feet = Length(u.feet)
foot = Length(u.foot)
ft = Length(u.ft)
full_moon_cycle = Time(u.full_moon_cycle)
full_moon_cycles = Time(u.full_moon_cycles)
g = Mass(u.g)
gaussian_year = Time(u.gaussian_year)
gaussian_years = Time(u.gaussian_years)
gee = Acceleration(u.gee)
gees = Acceleration(u.gees)
gibibyte = Information(u.gibibyte)
gibibytes = Information(u.gibibytes)
gram = Mass(u.gram)
grams = Mass(u.grams)
gray = Gray(u.gray)
h = Time(u.h)
hbar = Action(u.hbar)
henry = Inductance(u.henry)
henrys = Inductance(u.henrys)
hertz = Frequency(u.hertz)
hour = Time(u.hour)
hours = Time(u.hours)
hz = Frequency(u.hz)
inch = Length(u.inch)
inches = Length(u.inches)
joule = Energy(u.joule)
joules = Energy(u.joules)
julian_year = Time(u.julian_year)
julian_years = Time(u.julian_years)
kPa = EnergyDensity(u.kPa)
kat = Katal(u.kat)
katal = Katal(u.katal)
kelvin = Temperature(u.kelvin)
kelvins = Temperature(u.kelvins)
kg = Mass(u.kg)
kibibyte = Information(u.kibibyte)
kibibytes = Information(u.kibibytes)
kilogram = Mass(u.kilogram)
kilograms = Mass(u.kilograms)
kilometer = Length(u.kilometer)
kilometers = Length(u.kilometers)
kilopascal = EnergyDensity(u.kilopascal)
km = Length(u.km)
l = Volume(u.l)
lightyear = Length(u.lightyear)
lightyears = Length(u.lightyears)
liter = Volume(u.liter)
liters = Volume(u.liters)
lux = Illuminance(u.lux)
lx = Illuminance(u.lx)
ly = Length(u.ly)
m = Length(u.m)
mebibyte = Information(u.mebibyte)
mebibytes = Information(u.mebibytes)
meter = Length(u.meter)
meters = Length(u.meters)
mg = Mass(u.mg)
mho = Conductance(u.mho)
mhos = Conductance(u.mhos)
mi = Length(u.mi)
microgram = Mass(u.microgram)
micrograms = Mass(u.micrograms)
micrometer = Length(u.micrometer)
micrometers = Length(u.micrometers)
micron = Length(u.micron)
microns = Length(u.microns)
microsecond = Time(u.microsecond)
microseconds = Time(u.microseconds)
mil = AngularMil(u.mil)
mile = Length(u.mile)
miles = Length(u.miles)
milli_mass_unit = Mass(u.milli_mass_unit)
milligram = Mass(u.milligram)
milligrams = Mass(u.milligrams)
milliliter = Volume(u.milliliter)
milliliters = Volume(u.milliliters)
millimeter = Length(u.millimeter)
millimeters = Length(u.millimeters)
millisecond = Time(u.millisecond)
milliseconds = Time(u.milliseconds)
minute = Time(u.minute)
minutes = Time(u.minutes)
ml = Volume(u.ml)
mm = Length(u.mm)
mmHg = EnergyDensity(u.mmHg)
mmu = Mass(u.mmu)
mmus = Mass(u.mmus)
mol = AmountOfSubstance(u.mol)
mole = AmountOfSubstance(u.mole)
moles = AmountOfSubstance(u.moles)
ms = Time(u.ms)
nanometer = Length(u.nanometer)
nanometers = Length(u.nanometers)
nanosecond = Time(u.nanosecond)
nanoseconds = Time(u.nanoseconds)
nautical_mile = Length(u.nautical_mile)
nautical_miles = Length(u.nautical_miles)
newton = Force(u.newton)
newtons = Force(u.newtons)
nm = Length(u.nm)
nmi = Length(u.nmi)
ns = Time(u.ns)
ohm = Impedance(u.ohm)
ohms = Impedance(u.ohms)
optical_power = Dioptre(u.optical_power)
pa = EnergyDensity(u.pa)
pascal = EnergyDensity(u.pascal)
pascals = EnergyDensity(u.pascals)
pebibyte = Information(u.pebibyte)
pebibytes = Information(u.pebibytes)
percent = Dimensionless(u.percent)
percents = Dimensionless(u.percents)
permille = Dimensionless(u.permille)
picometer = Length(u.picometer)
picometers = Length(u.picometers)
picosecond = Time(u.picosecond)
picoseconds = Time(u.picoseconds)
planck = Action(u.planck)
planck_acceleration = Acceleration(u.planck_acceleration)
planck_angular_frequency = Frequency(u.planck_angular_frequency)
planck_area = Area(u.planck_area)
planck_charge = Charge(u.planck_charge)
planck_current = Current(u.planck_current)
planck_density = Density(u.planck_density)
planck_energy = Energy(u.planck_energy)
planck_energy_density = EnergyDensity(u.planck_energy_density)
planck_force = Force(u.planck_force)
planck_impedance = Impedance(u.planck_impedance)
planck_intensity = PlanckIntensity(u.planck_intensity)
planck_length = Length(u.planck_length)
planck_mass = Mass(u.planck_mass)
planck_momentum = Momentum(u.planck_momentum)
planck_power = Power(u.planck_power)
planck_pressure = EnergyDensity(u.planck_pressure)
planck_temperature = Temperature(u.planck_temperature)
planck_time = Time(u.planck_time)
planck_voltage = Voltage(u.planck_voltage)
planck_volume = Volume(u.planck_volume)
pm = Length(u.pm)
pound = Mass(u.pound)
pounds = Mass(u.pounds)
ps = Time(u.ps)
psi = EnergyDensity(u.psi)
quart = Volume(u.quart)
quarts = Volume(u.quarts)
rad = Angle(u.rad)
radian = Angle(u.radian)
radians = Angle(u.radians)
rutherford = Frequency(u.rutherford)
s = Time(u.s)
second = Time(u.second)
seconds = Time(u.seconds)
sidereal_year = Time(u.sidereal_year)
sidereal_years = Time(u.sidereal_years)
siemens = Conductance(u.siemens)
speed_of_light = Velocity(u.speed_of_light)
sr = SteRadian(u.sr)
steradian = SteRadian(u.steradian)
steradians = SteRadian(u.steradians)
tebibyte = Information(u.tebibyte)
tebibytes = Information(u.tebibytes)
tesla = MagneticDensity(u.tesla)
teslas = MagneticDensity(u.teslas)
torr = EnergyDensity(u.torr)
tropical_year = Time(u.tropical_year)
tropical_years = Time(u.tropical_years)
ug = Mass(u.ug)
um = Length(u.um)
us = Time(u.us)
v = Voltage(u.v)
vacuum_impedance = Impedance(u.vacuum_impedance)
vacuum_permittivity = Permittivity(u.vacuum_permittivity)
volt = Voltage(u.volt)
volts = Voltage(u.volts)
watt = Power(u.watt)
watts = Power(u.watts)
wb = MagneticFlux(u.wb)
weber = MagneticFlux(u.weber)
webers = MagneticFlux(u.webers)
yard = Length(u.yard)
yards = Length(u.yards)
yd = Length(u.yd)
year = Time(u.year)
years = Time(u.years)